/**
 * @file ProcessAudio_F32.h
 * @author Corey McCall (corey.mccall@femtosense.ai)
 * @brief This audio streaming and processing architecture is based on the Tympan example sketch "MyAudioAlgorithm"
 * @version 2.0
 * @date 2024-03-21
 *
 * @copyright Copyright (c) 2024
 *
 */

#ifndef PROCESSAUDIO_F32_H
#define PROCESSAUDIO_F32_H

#include "AINR.h"
#include <AudioStream.h>  //Audio streaming library

class ProcessAudio_F32 : public AudioStream_F32 {
public:

  //constructor
  ProcessAudio_F32(const AudioSettings_F32 &settings)
    : AudioStream_F32(1, inputQueueArray_f32) {
    ;
  };

  //This is called every time a new chunk of audio is received
  void update(void) {

    //get audio block
    audio_block_f32_t *audio_block_l;  //Left channel
    audio_block_f32_t *audio_block_r;  //Right channel

    audio_block_l = AudioStream_F32::receiveWritable_f32();
    audio_block_r = AudioStream_F32::allocate_f32();  // This grabs the right channel audio

    //return if L or R channel data is not ready yet
    if (!audio_block_r) {
      if (audio_block_l) AudioStream_F32::release((audio_block_f32_t *)audio_block_l);
      return;
    }
    if (!audio_block_l) {
      if (audio_block_r) AudioStream_F32::release((audio_block_f32_t *)audio_block_r);
      return;
    }

    //Process audio. TC2 processing supports core gating, while MP processing is simpler.
    if (isMP)
      processSpuMP(audio_block_l, audio_block_r);
    else
      processSpuTC2(audio_block_l, audio_block_r);

    ///transmit the block and release memory
    AudioStream_F32::transmit(audio_block_l, 0); //mono output
    AudioStream_F32::release(audio_block_r);
    AudioStream_F32::release(audio_block_l);
  }

  //process on SPU MP chip
  void processSpuMP(audio_block_f32_t *audio_block_l, audio_block_f32_t *audio_block_r) {

    //only process audio with AINR algorithm if not in loopback mode
    if (!loopback_mode) {

      SpuData audioSamples[2 * HOP_SIZE];
      for (int i = 0; i < HOP_SIZE; i++) {
        //get ready to send Left channel audio
        audioSamples[i] = audio_block_l->data[i] * 32767;  //convert to integer
      }

      if (CONFIRM_2MIC == 1) {
        for (int i = HOP_SIZE; i < (HOP_SIZE * 2); i++) {
          //get ready to send Right channel audio
          audioSamples[i] = audio_block_r->data[i - HOP_SIZE] * 32767;  //convert to integer
        }
      }

      //turn on cores
      SpuEnableCores(1, (DISABLE_CORE1 == 0));

      //Send most of the data to SPU, then increase clock speed and send the rest
      int writeDataLength = (CONFIRM_2MIC == 1) ? (2 * HOP_SIZE) : HOP_SIZE;
      SpuWriteData(audioSamples, writeDataLength - 4, 1, 0);  //dont start processing yet

      //speed up clock to process. You need to do this before the SPU processing starts
      if (CLKOD != CLKOD_IO)
        SpuClockSetOutputDivider(CLKOD);

      //send the rest of the audio to SPU
      SpuWriteData(&audioSamples[writeDataLength - 4], 4, 0, 1);

      //start execution timer
      unsigned long timer_extime = micros();

      //wait for interrupt to pull data back from SPU
      unsigned long intTimer = micros();
      SpuResult interruptValue = kSpuNotReady;
      while (interruptValue != kSpuSuccess) {
        interruptValue = SpuPollInterrupt();

        if ((micros() - intTimer) > (10.0 * HOP_SIZE / SAMPLE_RATE_HZ * 1000000)) {  //If no interrupt is seen after 10 hops, there was a problem, exit
          Serial.print(millis());
          Serial.println(": No data ready after several hops time, exiting AINR.");
          setLoopbackMode(1);
          return;
        }
      }

      //Print processing time
      if (PRINT_EXECUTION_TIME == 1) {
        Serial.println(micros() - timer_extime);
      }

      //make sure processing is done. This may not be needed.
      delayMicroseconds(10);

      //processing is done now, sleep to save power
      if (CLKOD != CLKOD_IO)
        SpuClockSetOutputDivider(CLKOD_IO);


      //read response from SPU
      SpuReadData(audioSamples, HOP_SIZE);

      //IO is done now, turn off cores to save power
      SpuEnableCores(0, 0);

      //left channel is used as mono audio output
      for (int i = 0; i < HOP_SIZE; i++) {
        audio_block_l->data[i] = 0.0000305185 * audioSamples[i];
      }
    }
  }

  //process on SPU TC2 chip
  void processSpuTC2(audio_block_f32_t *audio_block_l, audio_block_f32_t *audio_block_r) {

    //only process audio with AINR algorithm if not in loopback mode
    if (!loopback_mode) {

      SpuData audioSamples[2 * HOP_SIZE];
      for (int i = 0; i < HOP_SIZE; i++) {
        //get ready to send Left channel audio
        audioSamples[i] = audio_block_l->data[i] * 32767;  //convert to integer
      }

      if (CONFIRM_2MIC == 1) {
        for (int i = HOP_SIZE; i < HOP_SIZE * 2; i++) {
          //get ready to send Right channel audio
          audioSamples[i] = audio_block_r->data[i - HOP_SIZE] * 32767;  //convert to integer
        }
      }

      if (CORE_GATING_EN) {
        SpuEnableCores(1, 0);  //turn on core0, leave core1 off for now
      } else {
        SpuEnableCores(1, (DISABLE_CORE1 == 0));  //turn both cores on
      }

      //Send most of the data to SPU, then increase clock speed and send the rest
      int writeDataLength = (CONFIRM_2MIC == 1) ? (2 * HOP_SIZE) : HOP_SIZE;
      SpuWriteData(audioSamples, writeDataLength - 4, 1, 0);  //dont start processing yet

      //speed up clock to process. You need to do this before the SPU processing starts
      if (CLKOD != CLKOD_IO)
        SpuClockSetOutputDivider(CLKOD);

      //send the rest of the audio to SPU
      SpuWriteData(&audioSamples[writeDataLength - 4], 4, 0, 1);

      //start execution timer
      unsigned long timer_extime = micros();

      if (CORE_GATING_EN) {
        delayMicroseconds(CORE0_READY_US - SAFETY_MARGIN_US);  //wait for core0 to finish processing
      }

      //turn on core1. If not using core gating, this is actually already on.
      SpuEnableCores(1, (DISABLE_CORE1 == 0));

      if (CORE_GATING_EN) {
        delayMicroseconds(CORE_TRANSFER_US + SAFETY_MARGIN_US * 2);  //wait for core0-to-core1 transfer to finish
      }

      if (CORE_GATING_EN) {
        SpuEnableCores(0, 1);  //turn off core0, leave core1 on
      } else {
        SpuEnableCores(1, (DISABLE_CORE1 == 0));
      }

      //use delay to save IO power polling intrerupt on TC2
      if (CORE_GATING_EN) {
        delayMicroseconds(TOTAL_WAIT_TIME_US - (CORE0_READY_US - SAFETY_MARGIN_US) - (CORE_TRANSFER_US + SAFETY_MARGIN_US * 2));
      } else {
        delayMicroseconds(TOTAL_WAIT_TIME_US);
      }

      //wait for interrupt to pull data back from SPU
      unsigned long intTimer = micros();
      SpuResult interruptValue = kSpuNotReady;
      while (interruptValue != kSpuSuccess) {
        interruptValue = SpuPollInterrupt();

        if ((micros() - intTimer) > (10.0 * HOP_SIZE / SAMPLE_RATE_HZ * 1000000)) {  //If no interrupt is seen after 10 hops, there was a problem, exit
          Serial.print(millis());
          Serial.println(": No data ready after several hops time, exiting AINR.");
          setLoopbackMode(1);
          return;
        }
      }

      //Print processing time
      if (PRINT_EXECUTION_TIME == 1) {
        Serial.println(micros() - timer_extime);
      }

      //make sure processing is done. This may not be needed.
      delayMicroseconds(10);

      //processing is done now, sleep to save power
      if (CLKOD != CLKOD_IO)
        SpuClockSetOutputDivider(CLKOD_IO);

      //read response from SPU
      SpuReadData(audioSamples, HOP_SIZE);

      //IO is done now, turn off cores to save power
      SpuEnableCores(0, 0);

      //left channel is used as mono audio output
      for (int i = 0; i < HOP_SIZE; i++) {
        audio_block_l->data[i] = 0.0000305185 * audioSamples[i];
      }
    }
  }

  //sets loopbackMode tracker
  void setLoopbackMode(int setting) {
    loopback_mode = setting;
  }

  //gets loopbackMoode tracker
  int getLoopbackMode() {
    return loopback_mode;
  }

  //sets SPU architecture setting
  void setMP(int setting) {
    isMP = setting;
  }

private:
  audio_block_f32_t *inputQueueArray_f32[1];  //memory pointer for the input to this module

  int loopback_mode = 1;  //keep track of loopbackMode

  int isMP = 0;  //SPU architecture flag
};

#endif
