/**
 * @file spu001.h
 * @author Mathieu Le Goc (mathieu.legoc@femtosense.ai)
 * @brief
 * @version 1.1.2
 * @date 2024-03-18
 *
 * @copyright Copyright (c) 2024
 *
 */
#ifndef SPU001_H
#define SPU001_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>
#include <stdint.h>
#include <string.h>

#define SPU_SPI_FULL_DUPLEX 0
#define SPU_BUFFER_MAX_SIZE 256
#define kSpuPaddingTxLen 2
#define kSpuPaddingRxLen 4

#define kSpuMailboxIdValid 0
#define kSpuCoreIdValid 0x1F

#define MIN(x, y)           \
    ({                      \
        typeof(x) _x = (x); \
        typeof(y) _y = (y); \
        _x < _y ? _x : _y;  \
    })

typedef int16_t SpuData;
typedef uint32_t SpuAddress;
typedef uint32_t SpuModelData;
typedef uint32_t SpuConfigData;

typedef enum {
    kSpuVersionNone = 0,
    kSpuVersionTc2 = 1,
    kSpuVersionMp = 2,
} SpuVersion;

typedef enum {
    kSpuAddrSpiConf = (SpuConfigData)0x00,
    kSpuAddrIoConf = (SpuConfigData)0x04,
    kSpuAddrPllConf0 = (SpuConfigData)0x08,
    kSpuAddrPllConf1 = (SpuConfigData)0x0C,
    kSpuAddrPllConf2 = (SpuConfigData)0x10,
    kSpuAddrPllConf3 = (SpuConfigData)0x14,
    kSpuAddrPllConf4 = (SpuConfigData)0x18,
    kSpuAddrPllLock = (SpuConfigData)0x1C,
    kSpuAddrInterrupt = (SpuConfigData)0x24,
    kSpuAddrAudio = (SpuConfigData)0x00,
} SpuRegAddress;

typedef enum {
    kSpiIoReg = 0x00,
    kSpiApb = 0x01,
    kSpiAxis = 0x02,
    kSpiAxisLast = 0x03,
} SpuSpiInstruction;

typedef enum {
    kSpuSpiRead = 0x00,
    kSpuSpiWrite = 0x01,
} SpuSpiCommand;

typedef enum {
    kSpuSuccess = 0,
    kSpuError = 1,
    kSpuNotDetected = 2,
    kSpuConfigError = 3,
    kSpuFrameError = 4,
    kSpuBadParameters = 5,
    kSpuNotReady = 6,
    kSpuTestFailed = 7,
    kSpuTimeout = 8,
    kSpuPllConfigError = 9,
} SpuResult;

typedef enum {
    kSpuRegIoConfNoOsc = (SpuConfigData)0x0048,
    kSpuRegIoConfExtClk = (SpuConfigData)0x0248,
    kSpuRegIoConfOsc = (SpuConfigData)0x0448,
    kSpuRegIoIntEnable = (SpuConfigData)1 << 11,
    kSpuRegIoIntDisable = (SpuConfigData)0 << 11,
} SpuRegIoConfValues;

typedef enum {
    kSpuRegSpiConfCore0Disable = (SpuConfigData)1 << 1,
    kSpuRegSpiConfCore1Disable = (SpuConfigData)1 << 2,
    kSpuRegSpiConfCoresReset = (SpuConfigData)1 << 11,
} SpuRegSpiConfValues;

typedef enum {
    kSpuRegPll0ConfClockEnable = (SpuConfigData)1 << 6,
} SpuRegPll0ConfValues;

typedef enum {
    kSpuOsc = 0,     // Clock generated using SPU's oscillator
    kSpuExtClk = 1,  // Clock provided by the MCU on SPU's pin (OSC_PAD)
    kOscNone = 0xFF,
} SpuClockType;

typedef enum {
    kSpuInterruptRegister = 0,
    kSpuInterruptPin = 1,
    kSpuInterruptNone = 0xFF,
} SpuInterruptType;

typedef struct {
    SpuClockType clock_type;
    SpuInterruptType interrupt_type;
    uint16_t core_clock_frequency_mhz;
    uint8_t spi_clock_frequency_mhz;
    uint32_t crystal_frequency_hz;
} SpuConfiguration;

typedef union {
    struct ClockRegisters {
        SpuConfigData bwadj;
        SpuConfigData clkod;
        SpuConfigData clkf;
        SpuConfigData clkr;
    } registers;
    SpuConfigData data[sizeof(struct ClockRegisters) / sizeof(SpuConfigData)];
} SpuClockRegisters;

typedef struct {
    uint8_t command;
    SpuAddress address_start;
    SpuAddress address_end;
} __attribute__((packed)) SpuHeader;

typedef struct {
    uint32_t core_id;
    uint32_t program_counter;
} SpuInputParams;

typedef struct {
    uint32_t core_id;
    uint32_t mailbox_id;
} SpuOutputParams;

typedef union {
    struct ConfigTxMessage {
        SpuHeader header;
        SpuConfigData payload[SPU_BUFFER_MAX_SIZE];
        uint16_t _padding;
    } __attribute__((packed)) message;
    uint8_t data[sizeof(struct ConfigTxMessage)];
} SpuConfigTxMessage;

typedef union {
    struct ConfigRxMessage {
        SpuHeader header;
        uint32_t _padding;
        SpuConfigData payload[SPU_BUFFER_MAX_SIZE];
    } __attribute__((packed)) message;
    uint8_t data[sizeof(struct ConfigRxMessage)];
} SpuConfigRxMessage;

typedef union {
    struct DataRxMessage {
        SpuHeader header;
        uint32_t _padding;
        SpuOutputParams params;
        SpuData payload[SPU_BUFFER_MAX_SIZE];
    } __attribute__((packed)) message;
    uint8_t data[sizeof(struct DataRxMessage)];
} SpuDataRxMessage;

typedef union {
    struct DataTxMessage {
        SpuHeader header;
        SpuInputParams params;
        SpuData payload[SPU_BUFFER_MAX_SIZE];
        uint16_t _padding;
    } __attribute__((packed)) message;
    uint8_t data[sizeof(struct DataTxMessage)];
} SpuDataTxMessage;

typedef union {
    struct DataTxMessageContinued {
        SpuHeader header;
        SpuData payload[SPU_BUFFER_MAX_SIZE];
        uint16_t _padding;
    } __attribute__((packed)) message;
    uint8_t data[sizeof(struct DataTxMessageContinued)];
} SpuDataTxMessageContinued;

SpuResult SpuInit(SpuConfiguration *configuration);
SpuResult SpuReset(void);

SpuResult SpuEnableCores(bool core_0, bool core_1);
SpuResult SpuEnablePll(SpuClockType clk_source);
SpuResult SpuDisablePll(void);

SpuResult SpuWriteModelValue(SpuAddress address, SpuModelData data);
SpuResult SpuWriteModelBatch(SpuAddress address, SpuModelData *data, uint16_t length);

SpuResult SpuReadData(SpuData *data, uint16_t length);
SpuResult SpuWriteData(SpuData *data, uint16_t length, bool first_message, bool last_message);

SpuResult SpuEnterSleepMode(void);
SpuResult SpuExitSleepMode(void);

SpuResult SpuPollInterrupt(void);

const char *SpuResultToString(SpuResult error);
SpuVersion SpuGetHardwareVersion(void);

SpuResult SpuClockSetOutputDivider(uint8_t divider);
#ifdef __cplusplus
}
#endif

#endif /* SPU001_H */
