/**
 * @file SPEECH_COMMANDS.ino
 * @author Corey McCall (corey.mccall@femtosense.ai)
 * @brief Speech Commands Example 
 * @version 2.0
 * @date 2024-03-21
 *
 * @copyright Copyright (c) 2024
 *
 * Speech Commands
 * This firmware programs the SPU with a variation of a speech commands algorithm that listens for commands. In multi-command models, 
 * the first command represents silence, and up to 99 other commands can be set in the XPROG_P file.
 * When a command is heard, the red and green LEDs flash the ID number of the command.
 * 
 * The command is also displayed on the serial terminal or GUI along with the confidence metric. Various debugging information is also displayed in 
 * the serial terminal.
 *
 * Before running this firmware, do the following:
 * 1) Insert an SD card with the files 0PROG_A and 0PROG_D corresponding to the example model. 0PROG_P can optionally be included
 *    to change parameters from the default.
 * 2) Make sure the EVK is connected according to the quick start guide.
 * 3) Ensure that AUDIO_BLOCK_SAMPLES in AudioStream.h is set to at least the value of HOP_SIZE*2 in spu.h below. This file is likely located at:
 * C:\Users\<USERNAME>\AppData\Local\Arduino15\packages\teensy\hardware\avr\1.57.2\cores\teensy4\AudioStream.h or similar
 * 4) The Button2 library by Lennart Hennigs, Teensy 1.57.2, and the Tympan library are required to run this firmware. Install them in 
 * the Arduino Board and Library Managers.
 *
 * Pressing the button toggles power down mode so that the static power of the SPU can be measured (memory is retained, but no processing is being done)
 *
 */

#include <Tympan_Library.h>             //Tympan library
#include <SD.h>                         //SD card library
#include <Button2.h>                    //Button library
#include <SPI.h>                        //SPI library
#include "spu001.h"                     //Femtosense SPU library
#include "ProcessAudio_F32.h"           //Audio processing algorithm
#include "AudioEffectLowpass_FD_F32.h"  //Digital low pass filter from Tympan audio library
#include "SpeechCommands.h"

/*       Default parameters in the section below will be overwritten by XPROG_P file if it exists.        */
/*********************************************************************************************************/
//Firmware parameters:
int PRINT_EXECUTION_TIME = 0;        //prints SPU processing timer when value is 1
float INPUT_GAIN_DB = 30.0f;         //gain of the microphone (0-47.5dB in 0.5dB steps)
float OUTPUT_MAX_GAIN_DB = 0.0f;     //max gain of output
int USE_EXTERNAL_MIC = 0;            //use external mic (0=no, 1=mic jack,2=header)
String CONFIRM_ARCHITECTURE = "MP";  //confirm that actual SPU architecture matches model
String CONFIRM_MODEL_TYPE = "SC";    //confirm that model type matches firmware
int CONFIRM_HOP_SIZE = HOP_SIZE;     //confirm that the model hop size matches the firmware
int CONFIRM_2MIC = 0;                //use stereo mic input instead of mono
int ENABLE_GUI = 0;                  //send serial data in Femtosense GUI format

//Power tuning parameters:
int VCO_MHZ = 100;       //SPU clock speed in MHz
int CLKOD = 1;           //SPU Clock Output Divider (must be even or 1, and <= 16)
int USE_MEMORY_FSM = 0;  //Use FSM memory mode when value is 1
int DISABLE_CORE1 = 0;   //Leave second core off

//Power tuning parameters (TC2 only):
int CORE_GATING_EN = 0;               //Enable core gating when value is 1
unsigned int CORE0_READY_US = 0;      //Core 0 processing time in microseconds
unsigned int CORE_TRANSFER_US = 0;    //Core-to-core transfer time in microseconds
unsigned int SAFETY_MARGIN_US = 0;    //Safety margin around core-to-core transfer to account for inconsistant execution time (in microseconds)
unsigned int TOTAL_WAIT_TIME_US = 0;  //total time to wait for processing before checking interrupt

//Other SPU parameters:
uint32_t SPI_SPEED = 16000000;  //SPU SPI speed
int USE_ONBOARD_XTAL = 1;       //use oscillator configuration instead of reference clock when value is 1 (set jumper)
int USE_INT_PIN = 1;            //use SPU interrupt pin instead of register when value is 1 (set jumper)

//calculated parameters
int CLKOD_IO = VCO_MHZ / (2 * SPI_SPEED / 1000000);  //SPU Clock Output Divider during IO

//ML model parameters
String speechCommands[99] = { "<Unknown Command> Configure command names in 0PROG_P file." };  //speech command names. Does not include class 0.
int NUMBER_OF_COMMANDS = 1;                                                                    //does not include class 0 (silence)

/*********************************************************************************************************/

//configure audio
//NOTE: There is a bug in the Tympan library with 8KHz audio. To use 8KHz, sample at 16KHz and decimate/anti-alias back to 8KHz.
AudioSettings_F32 audio_settings((SAMPLE_RATE_HZ == 8000) ? 16000 : SAMPLE_RATE_HZ, (SAMPLE_RATE_HZ == 8000) ? (HOP_SIZE * 2) : HOP_SIZE);

//create audio library objects for handling the audio
Tympan myTympan(TympanRev::E, audio_settings);           //Initialize Tympan
AudioInputI2S_F32 i2sIn(audio_settings);                 //Digital audio from Codec.
ProcessAudio_F32 processSpeechCommands(audio_settings);  //Attach audio processing algorithm in ProcessAudio_F32.h. This runs every hop.
AudioOutputI2S_F32 i2sOut(audio_settings);               //Digital audio to Codec.
AudioEffectLowpass_FD_F32 AAFilter(audio_settings);      //Anti-aliasing filter

//audio connections
AudioConnection_F32 patchCord1(i2sIn, AAFilter);                  //connect both inputs to the AA filter
AudioConnection_F32 patchCord2(AAFilter, processSpeechCommands);  //connect to the processing algorithm

//connect the input audio to the headphone jack so you can hear what the algorithm hears.
AudioConnection_F32 patchCord3(AAFilter, 0, i2sOut, 0);
AudioConnection_F32 patchCord4(AAFilter, 0, i2sOut, 1);

//setup button
Button2 button;

//current powerdown state
int powerDownMode = 1;  // 1=SPU off, 0=SPU active. Start in active mode.

//state variables for non-blocking blink notification
unsigned long blinkTimer = 0;
int blinksRemaining = 0;
int blinkState = 0;

//notification blinking speed period
const unsigned long blinkInterval = 150; 

//runs once at startup
void setup() {

  //Initialize IO
  IOSetup();

  Serial.println("Femtosense Speech Commands Example v2.0");

  //Mount SD card
  if (!SD.begin(BUILTIN_SDCARD)) {
    error("SD card initialization failed! Check that card is inserted correctly.", 4, 1);
  }

  //select file set. Up to 9 additional models can be used as alternates
  int fileSet = getFileSet();
  Serial.println("Fileset: " + String(fileSet) + " selected.");

  //Set up firmware with model parameters
  if (!setModelParameters(fileSet)) {
    error("Model parameters on SD card are invalid", 4, 1);
  }

  //Initialize SPU and check that everything looks right
  startSPU();

  //Program SPU from SD Card
  SpuClockSetOutputDivider(CLKOD_IO);  //set appropriate clocking for IO transfers
  if (!programModelBurst(SPU_BUFFER_MAX_SIZE, fileSet))  
    error("Model could not be loaded from SD card.", 4, 1);

  //disable cores until needed
  SpuEnableCores(0, 0);

  //set chip architecture so processing function knows what sequence to use
  processSpeechCommands.setMP(CONFIRM_ARCHITECTURE == "MP");

  //Start audio feed
  AudioMemory_F32(20, audio_settings);                 //allocate the audio memory
  int N_FFT = audio_settings.audio_block_samples * 2;  //setup AA filter (only really neccessary for 8khz audio)
  AAFilter.setup(audio_settings, N_FFT);
  AAFilter.setCutoff_Hz(SAMPLE_RATE_HZ / 2);
  myTympan.enable();  //activate Codec

  //select microphone input
  if (USE_EXTERNAL_MIC == 1) {
    myTympan.inputSelect(TYMPAN_INPUT_JACK_AS_MIC);
    Serial.println("Connected to External Mic Jack!");
  } else if (USE_EXTERNAL_MIC == 2) {
    myTympan.inputSelect(TYMPAN_INPUT_LINE_IN);
    Serial.println("Connected to External Mic Header!");
  } else {
    myTympan.inputSelect(TYMPAN_INPUT_ON_BOARD_MIC);
  }
  myTympan.setInputGain_dB(INPUT_GAIN_DB);  //set input gain
  myTympan.volume_dB(-6);                  //set output driver gain to a resonable level for listening.

  //enable button
  button.begin(BUTTON_SW1);
  button.setTapHandler(tap);

  //activate processing
  powerDownMode = 0;
  updateMode();

  //Flash LED green to indicate device is ready
  flashLED(LED_GREEN, 3);
  Serial.println("Setup complete.");
}

//runs after setup and repeats forever.
void loop() {

  //check and see if processing loop exited because of an error
  if (powerDownMode != processSpeechCommands.getPowerDownMode()) {

    //go to power down mode and permanently error.
    powerDownMode = powerDownMode ? 0 : 1;
    updateMode();
    error("SPU processing error, restart required. Check power optimization parameters and PLL frequency. If this error occured on first frame, check that the model file is correct and use XPROG_P.", 6, 1);
  }

  //handle result notification
  handleResultNotification();

  //handle button press
  button.loop();  //if button is not working, try increasing SPU clock speed so that interrupt asserts sooner.
}

//Use button to select which fileset to use. Pressing the button moves selection to next fileset (1PROG-9PROG),
//and holding the button selects the current fileset. A blink confirms.
//Returns fileset id
int getFileSet() {
  //If button is not pressed, use default (0PROG_X)
  if (digitalRead(BUTTON_SW1) == HIGH)
    return 0;

  //turn on LEDs to show selection mode until button is released
  Serial.println("Selecting fileset...Hold button to confirm selection");
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  while (digitalRead(BUTTON_SW1) == LOW)
    ;
  digitalWrite(LED_RED, LOW);
  digitalWrite(LED_GREEN, LOW);
  delay(250);

  //check for file numbers 1-9 (check A files only, D and P will be validated when the files are loaded)
  bool fileSetAvailable[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  bool filesFound = 0;
  for (int i = 1; i <= 9; i++) {
    String addressFileName = String(i) + "PROG_A";
    char addressFileNameChar[12];
    addressFileName.toCharArray(addressFileNameChar, 12);
    if (SD.exists(addressFileNameChar)) {
      fileSetAvailable[i - 1] = 1;
      filesFound = 1;
    } else {
      addressFileName += ".txt";
      addressFileName.toCharArray(addressFileNameChar, 12);
      if (SD.exists(addressFileNameChar)) {
        fileSetAvailable[i - 1] = 1;
        filesFound = 1;
      }
    }
  }

  //no alternative files found
  if (!filesFound)
    return 0;

  //If files are found, get selection from user.
  int currentSet = 1;
  while (1) {
    //if file is available
    if (fileSetAvailable[currentSet - 1] == 1) {
      Serial.println(String(currentSet) + "?");
      flashLEDs(currentSet);
      //wait for button selection
      while (digitalRead(BUTTON_SW1) == HIGH)
        ;
      delay(10);  //debounce
      //hold button for 1s or continue to next selection
      unsigned long timer = millis();
      while (digitalRead(BUTTON_SW1) == LOW) {
        if ((millis() - timer) > 1000) {
          //confirm and return
          flashLEDs(currentSet);
          return currentSet;
        }
      }

      //if button was not held, select next option
      delay(100);  //debounce
    }

    //loop back around
    if (currentSet == 10)
      currentSet = 1;
    else
      currentSet++;
  }
}

//read one line from a 0PROG file
uint32_t read0PROGLine(File& PROGFile, byte CRLF) {

  //string for line of data
  char progLine[9];
  progLine[8] = '\0';

  //read value
  PROGFile.readBytes(progLine, 8);
  uint32_t progValue = strtoul(progLine, '\0', 16);

  //consume line endings
  PROGFile.readBytes(progLine, 1);
  if (CRLF)
    PROGFile.readBytes(progLine, 1);

  return progValue;
}


//Reads files XPROG_A and XPROG_D from root directory of SD card and programs SPU
//maxBurstSize is the maximum number of values to send in a single SPI transaction
//fileset is the id of the XPROG files.
//Returns 1 if successful, else returns 0
int programModelBurst(int maxBurstSize, int fileSet) {

  //Setup files
  String addressFileName = String(fileSet) + "PROG_A";
  char addressFileNameChar[12];
  addressFileName.toCharArray(addressFileNameChar, 12);
  File addressFile = SD.open(addressFileNameChar);

  String dataFileName = String(fileSet) + "PROG_D";
  char dataFileNameChar[12];
  dataFileName.toCharArray(dataFileNameChar, 12);
  File dataFile = SD.open(dataFileNameChar);

  //make sure files exist
  if (!addressFile) {
    addressFileName += ".txt";
    addressFileName.toCharArray(addressFileNameChar, 12);
    addressFile = SD.open(addressFileNameChar);
    if (!addressFile) {
      Serial.println("Cannot find file PROG_A on SD card");
      return 0;
    }
  }
  if (!dataFile) {
    dataFileName += ".txt";
    dataFileName.toCharArray(dataFileNameChar, 12);
    dataFile = SD.open(dataFileNameChar);
    if (!dataFile) {
      Serial.println("Cannot find file PROG_D on SD card");
      return 0;
    }
  }

  //check for linebreak style
  char CRLF_CHECK[10];
  addressFile.readBytes(CRLF_CHECK, 10);
  byte CRLF_addressFile = (CRLF_CHECK[9] == '\012') ? 1 : 0;  //does address file have 2 line break characters (CR+LF)?
  dataFile.readBytes(CRLF_CHECK, 10);
  byte CRLF_dataFile = (CRLF_CHECK[9] == '\012') ? 1 : 0;  //does data file?

  //reset files after line break check
  addressFile.seek(0);
  dataFile.seek(0);

  //data buffer
  SpuModelData datas[maxBurstSize];

  //loop through file looking for sequential addresses to burst in one transaction
  SpuAddress currentAddress = (SpuAddress)read0PROGLine(addressFile, CRLF_addressFile);
  int line = 1;
  unsigned long timer = millis();
  do {
    //new chunk of data to send
    SpuAddress startAddress = currentAddress;
    int burstSize = 0;

    //look for sequential addresses
    SpuAddress prevAddress;
    do {
      prevAddress = currentAddress;
      if (addressFile.available()) {
        currentAddress = (SpuAddress)read0PROGLine(addressFile, CRLF_addressFile);
        line++;

        burstSize++;
      } else {
        break;
      }
    } while (((currentAddress - prevAddress) == 4) && (burstSize < maxBurstSize));

    //burst data
    Serial.print("Programming Model: ");
    Serial.println(line - 1);

    for (int i = 0; i < burstSize; i++) {
      datas[i] = (SpuModelData)read0PROGLine(dataFile, CRLF_dataFile);

      if (USE_MEMORY_FSM) {
        SpuAddress address = startAddress + i * 4;
        if ((address <= 0x98) && (address >= 0x34) && (datas[i] == 0x11098)) {
          datas[i] = 0x15098;
        }
      }
    }
    SpuWriteModelBatch(startAddress, datas, burstSize);

    //loop until end of file
  } while (addressFile.available());

  //send final line of data
  Serial.print("Programming Model: ");
  Serial.print(line);
  Serial.println(", done.");

  datas[0] = (SpuModelData)read0PROGLine(dataFile, CRLF_dataFile);
  SpuWriteModelValue(currentAddress, datas[0]);

  //completed successfully
  Serial.print("Loaded Model in ");
  Serial.print(millis() - timer);
  Serial.println("ms.");

  return 1;
}

//Flases LED flashCount number of times.
void flashLED(int LEDPin, int flashCount) {
  for (int i = 0; i < flashCount; i++) {
    digitalWrite(LEDPin, HIGH);
    delay(150);
    digitalWrite(LEDPin, LOW);
    delay(150);
  }
}

//Flases both LED flashCount number of times.
void flashLEDs(int flashCount) {
  for (int i = 0; i < flashCount; i++) {
    digitalWrite(LED_RED, HIGH);
    digitalWrite(LED_GREEN, HIGH);
    delay(150);
    digitalWrite(LED_RED, LOW);
    digitalWrite(LED_GREEN, LOW);
    delay(150);
  }
}

//Prints error message and ends code execution while flashing red LED flashCount number of times.
void error(String errorMessage, int flashCount, int stopProgram) {

  //print error message
  Serial.print("ERROR");
  Serial.print(flashCount);
  Serial.print(": ");
  Serial.println(errorMessage);

  //loop forever
  while (1) {
    flashLED(LED_RED, flashCount);
    delay(1000);

    //loop once
    if (!stopProgram)
      break;
  }
}

//Called whenever the button is pressed
void tap(Button2& button) {

  //toggle between power down mode and active mode
  powerDownMode = (powerDownMode) ? 0 : 1;

  updateMode();
}

//Toggles between power down mode and active mode. Call after powerDownMode is changed.
void updateMode() {

  //power down processing
  if (powerDownMode) {

    //stop flashing
    blinksRemaining = 0;
    digitalWrite(LED_GREEN, LOW);
    digitalWrite(LED_RED, LOW);

    //stop processing
    processSpeechCommands.setPowerDownMode(powerDownMode);

    //turn off clock and cores
    SpuDisablePll();

    Serial.println("Stopped listening for speech commands...");

    //turn on red LED
    digitalWrite(LED_RED, powerDownMode);

    //activate processing
  } else {
    //turn on clock
    SpuResult SPUError = SpuEnablePll((USE_ONBOARD_XTAL == 0) ? kSpuExtClk : kSpuOsc);
    if (SPUError != kSpuSuccess) {
      error("SPU failed to restart PLL: " + String(SpuResultToString(SPUError)), 3, 1);
    }

    Serial.println("Now listening for speech commands...");

    //turn off red LED
    digitalWrite(LED_RED, powerDownMode);

    //start AINR
    processSpeechCommands.setPowerDownMode(powerDownMode);
  }
}

//Reads file 0PROG_P from root directory of SD card and sets model parameters
//If the file does not exist, a non-blocking error is shown, and the default values are not changed.
//Returns 1 if successful and 0 if the parameters read are invalid.
int setModelParameters(int fileSet) {

  //load file
  String fileName = String(fileSet) + "PROG_P";
  char fileNameChar[12];
  fileName.toCharArray(fileNameChar, 12);

  File parametersFile = SD.open(fileNameChar);
  if (!parametersFile) {
    fileName += ".txt";
    fileName.toCharArray(fileNameChar, 12);
    parametersFile = SD.open(fileNameChar);
    if (!parametersFile) {
      error(String(fileSet) + "PROG_P file not found, using default parameters...", 1, 0);
      return 1;
    }
  }

  //0PROG_P found!
  flashLED(LED_GREEN, 1);

  //Keep track of core gating parameters. All must be set if CORE_GATING_EN == 1.
  int setCORE_GATING_EN = -1;
  int setCORE0_READY_US = -1;
  int setCORE_TRANSFER_US = -1;
  int setSAFETY_MARGIN_US = -1;

  while (parametersFile.available()) {

    //read space delimited file
    String parameterNameStr = parametersFile.readStringUntil(' ');
    String parameterValueStr = parametersFile.readStringUntil('\n');

    //set parameters and print parsed values
    String strval;
    Serial.print(String(fileSet) + String("PROG_P: "));
    if (parameterNameStr == "SPI_SPEED") {
      SPI_SPEED = parameterValueStr.toInt();
      strval = String(SPI_SPEED);
    } else if (parameterNameStr == "VCO_MHZ") {
      VCO_MHZ = parameterValueStr.toInt();
      strval = String(VCO_MHZ);
    } else if (parameterNameStr == "CORE_GATING_EN") {
      setCORE_GATING_EN = parameterValueStr.toInt();
      strval = String(setCORE_GATING_EN);
    } else if (parameterNameStr == "CORE0_READY_US") {
      setCORE0_READY_US = parameterValueStr.toInt();
      strval = String(setCORE0_READY_US);
    } else if (parameterNameStr == "CORE_TRANSFER_US") {
      setCORE_TRANSFER_US = parameterValueStr.toInt();
      strval = String(setCORE_TRANSFER_US);
    } else if (parameterNameStr == "SAFETY_MARGIN_US") {
      setSAFETY_MARGIN_US = parameterValueStr.toInt();
      strval = String(setSAFETY_MARGIN_US);
    } else if (parameterNameStr == "TOTAL_WAIT_TIME_US") {
      TOTAL_WAIT_TIME_US = parameterValueStr.toInt();
      strval = String(TOTAL_WAIT_TIME_US);
    } else if (parameterNameStr == "USE_MEMORY_FSM") {
      USE_MEMORY_FSM = parameterValueStr.toInt();
      strval = String(USE_MEMORY_FSM);
    } else if (parameterNameStr == "INPUT_GAIN_DB") {
      INPUT_GAIN_DB = parameterValueStr.toFloat();
      strval = String(INPUT_GAIN_DB);
    } else if (parameterNameStr == "CLKOD") {
      CLKOD = parameterValueStr.toInt();
      strval = String(CLKOD);
    } else if (parameterNameStr == "OUTPUT_MAX_GAIN_DB") {
      OUTPUT_MAX_GAIN_DB = parameterValueStr.toFloat();
      strval = String(OUTPUT_MAX_GAIN_DB);
    } else if (parameterNameStr == "PRINT_EXECUTION_TIME") {
      PRINT_EXECUTION_TIME = parameterValueStr.toInt();
      strval = String(PRINT_EXECUTION_TIME);
    } else if (parameterNameStr == "USE_EXTERNAL_MIC") {
      USE_EXTERNAL_MIC = parameterValueStr.toInt();
      strval = String(USE_EXTERNAL_MIC);
    } else if (parameterNameStr == "CONFIRM_HOP_SIZE") {
      CONFIRM_HOP_SIZE = parameterValueStr.toInt();
      strval = String(CONFIRM_HOP_SIZE);
    } else if (parameterNameStr == "CONFIRM_2MIC") {
      CONFIRM_2MIC = parameterValueStr.toInt();
      strval = String(CONFIRM_2MIC);
    } else if (parameterNameStr == "USE_ONBOARD_XTAL") {
      USE_ONBOARD_XTAL = parameterValueStr.toInt();
      strval = String(USE_ONBOARD_XTAL);
    } else if (parameterNameStr == "USE_INT_PIN") {
      USE_INT_PIN = parameterValueStr.toInt();
      strval = String(USE_INT_PIN);
    } else if (parameterNameStr == "ENABLE_GUI") {
      ENABLE_GUI = parameterValueStr.toInt();
      strval = String(ENABLE_GUI);
    } else if (parameterNameStr == "DISABLE_CORE1") {
      DISABLE_CORE1 = parameterValueStr.toInt();
      strval = String(DISABLE_CORE1);
    } else if (parameterNameStr == "CONFIRM_ARCHITECTURE") {
      CONFIRM_ARCHITECTURE = parameterValueStr.trim();
      strval = CONFIRM_ARCHITECTURE;
    } else if (parameterNameStr == "CONFIRM_MODEL_TYPE") {
      CONFIRM_MODEL_TYPE = parameterValueStr.trim();
      strval = CONFIRM_MODEL_TYPE;
    } else if (parameterNameStr == "COMMAND_NAME") {
      parameterValueStr = parameterValueStr.trim();

      //parse parameter value and name
      int commandIndex = parameterValueStr.substring(0, parameterValueStr.indexOf(' ')).toInt();
      String commandName = parameterValueStr.substring(parameterValueStr.indexOf(' ') + 1, parameterValueStr.length());
      if (commandIndex > 99) {
        Serial.println("Command Index " + String(commandIndex) + " is too large (>99)");
      }

      //save
      if (commandIndex > NUMBER_OF_COMMANDS)
        NUMBER_OF_COMMANDS = commandIndex;
      speechCommands[commandIndex - 1] = commandName;

      strval = String(commandIndex) + ":" + commandName;
    } else {
      strval = "<This is an unknown parameter!>";
    }
    Serial.print(parameterNameStr);
    Serial.print(" = ");
    Serial.println(strval);
  }

  //Check VCO
  if ((VCO_MHZ < 30) || (VCO_MHZ > 268)) {
    Serial.println("Power tuning parameters error-  VCO_MHZ must be: 30 <= VCO_MHZ <= 268.");
    return 0;
  }
  int SPI_SPEED_MHZ = SPI_SPEED / 1000000;
  if (VCO_MHZ < (2 * SPI_SPEED_MHZ)) {
    Serial.println("Power tuning parameters error-  VCO_MHZ must be: >= 2 * SPI_SPEED in MHz.");
    return 0;
  }

  //Check CLKOD
  if ((CLKOD < 1 || CLKOD > 16) || (CLKOD != 1 && CLKOD % 2 == 1)) {
    Serial.println("Power tuning parameters error-  CLKOD must be: 1 <= CLKOD <= 16 and either 1 or even.");
    return 0;
  }

  //check core gating parameters
  if ((setCORE_GATING_EN == 1)) {
    //check that all parameters needed are included
    if ((setCORE0_READY_US > -1) && (setCORE_TRANSFER_US > -1) && (setSAFETY_MARGIN_US > -1)) {
      //set core gating parameters
      CORE_GATING_EN = setCORE_GATING_EN;
      CORE0_READY_US = setCORE0_READY_US;
      CORE_TRANSFER_US = setCORE_TRANSFER_US;
      SAFETY_MARGIN_US = setSAFETY_MARGIN_US;

      //parameters invalid
      if (!((!CORE_GATING_EN) || (((CORE0_READY_US - SAFETY_MARGIN_US) + (CORE_TRANSFER_US + 2 * SAFETY_MARGIN_US)) <= TOTAL_WAIT_TIME_US))) {
        Serial.println("Power tuning parameters error- individual delays are greater than total wait time allowed");
        return 0;
      }

      Serial.println("Core gating enabled!");

      //parmeters invalid
    } else {
      Serial.println("Core gating is enabled, but not all required parameters were set.");
      return 0;
    }
  }

  //check that interrupt is enabled on TC2
  if ((CONFIRM_ARCHITECTURE == "TC2") && (USE_INT_PIN == 0)) {
    Serial.println("You may not disable the interrupt pin when using TC2.");
    return 0;
  }


  //reset CLKOD_IO. It will be validated in SpuClockSetOutputDivider()
  CLKOD_IO = VCO_MHZ / (2 * SPI_SPEED / 1000000);
  Serial.print("CLKOD_IO (calculated) = ");
  Serial.println(CLKOD_IO);

  //return 1 for success.
  return 1;
}

//Setup required IO
void IOSetup() {
  //Initialize serial port
  Serial.begin(115200);

  //setup button
  pinMode(BUTTON_SW1, INPUT);
  pinMode(BUTTON_SW2, INPUT);  //currently unused

  //Enable SPU pins
  pinMode(PSU_PIN, OUTPUT);  //PSU
  digitalWrite(PSU_PIN, HIGH);
  pinMode(SS_SPU, OUTPUT);  //SPI SS
  digitalWrite(SS_SPU, HIGH);
  SPI.begin();                     //SPI
  pinMode(SPU_RESET_PIN, OUTPUT);  //Reset
  digitalWrite(SPU_RESET_PIN, LOW);

  //Disable flash chip that shares SPI bus with SPU on EVB board
  pinMode(SS_FLASH, OUTPUT);
  digitalWrite(SS_FLASH, HIGH);

  //Enable reference clock
  if (USE_ONBOARD_XTAL == 0) {
    analogWriteFrequency(SPU_INT_OR_REFCLK, 32768);
    analogWrite(SPU_INT_OR_REFCLK, 128);
  }

  //enable interrupt
  if (USE_INT_PIN == 1) {
    //interrupt and reference clock cannot be used a the same time
    if (USE_ONBOARD_XTAL == 0) {
      error("When external reference clock is used, the interrupt cannot be used", 7, 0);
    } else {
      pinMode(SPU_INT_OR_REFCLK, INPUT);
    }
  }
}

//Check that everything looks okay and error out if not. If everything is okay, SPU will be started.
void startSPU() {
  //Check that AUDIO_BLOCK_SAMPLES is set correctly
  //This file is located at: C:\Users\<USERNAME>\AppData\Local\Arduino15\packages\teensy\hardware\avr\1.57.2\cores\teensy4\AudioStream.h or similar
  if (SAMPLE_RATE_HZ == 8000) {
    if (AUDIO_BLOCK_SAMPLES < HOP_SIZE * 2)
      error("You must increase AUDIO_BLOCK_SAMPLES in AudioStream.h to be >= HOP_SIZE * 2", 2, 1);
  } else {
    if (AUDIO_BLOCK_SAMPLES < HOP_SIZE)
      error("You must increase AUDIO_BLOCK_SAMPLES in AudioStream.h to be >= HOP_SIZE", 2, 1);
  }

  //Check that HOP_SIZE is set correctly
  if (CONFIRM_HOP_SIZE != HOP_SIZE)
    error("HOP_SIZE does not match model file. Update HOP_SIZE SpeechCommands.h to the correct value", 7, 1);

  //Check that model type is correct
  if (CONFIRM_MODEL_TYPE != "SC")
    error("This model is not an Speech Commands model. Use a Speech Commands model.", 7, 1);

  SpuConfiguration spuConfig = { .clock_type = (USE_ONBOARD_XTAL == 0) ? kSpuExtClk : kSpuOsc,
                                 .interrupt_type = USE_INT_PIN ? kSpuInterruptPin : kSpuInterruptRegister,
                                 .core_clock_frequency_mhz = (uint16_t)VCO_MHZ,
                                 .spi_clock_frequency_mhz = (uint8_t)(SPI_SPEED / 1000000),
                                 .crystal_frequency_hz = 32768 };

  SpuResult SPUError = SpuInit(&spuConfig);
  if (SPUError != kSpuSuccess) {
    error("SPU initialization failure: " + String(SpuResultToString(SPUError)), 3, 1);
  }

  //Check that architecture is set correct
  SpuVersion checkVersion = (CONFIRM_ARCHITECTURE == "MP") ? kSpuVersionMp : kSpuVersionTc2;
  if (checkVersion != SpuGetHardwareVersion()) {
    if (checkVersion == kSpuVersionMp) {
      error("This model is compiled for SPU-001 (MP), but you are using SPU-001 (TC2)! Please compile the correct model", 7, 1);
    } else {
      error("This model is compiled for SPU-001 (TC2), but you are using SPU-001 (MP)! Please compile the correct model", 7, 1);
    }
  }

  //Check that core gating is supported
  if ((CONFIRM_ARCHITECTURE == "MP") && (CORE_GATING_EN == 1))
    error("Core gating is only supported on SPU-001 (TC2), but you are using SPU-001 (MP)!", 7, 1);
}

//non-blocking blink notification
void handleResultNotification() {
  //check and reset notification
  int newCommand = processSpeechCommands.getResultNotification();
  processSpeechCommands.setResultNotification(0);

  //print results
  if (newCommand) {
    Serial.println(String(millis() / 1000.0) + "s: " + String(newCommand) + "-" + speechCommands[newCommand - 1]);
    //show GUI
    if (ENABLE_GUI == 1) {
      uint8_t GUIOutput[] = { '~', 0, '!', '\r', '\n' };
      GUIOutput[1] = newCommand + 50;
      Serial.write(GUIOutput, sizeof(GUIOutput));
    }
  }

  //blink results

  //start blink notification if not already busy
  if (!blinksRemaining && newCommand) {
    //setup blinking
    blinksRemaining = newCommand;
    blinkTimer = millis();
  }

  //no more blinking to do
  if (!blinksRemaining)
    return;

  //blink
  unsigned long currentTime = millis();

  //turn on
  if (((currentTime - blinkTimer) < blinkInterval) && (blinkState == 0)) {
    digitalWrite(LED_GREEN, HIGH);
    digitalWrite(LED_RED, HIGH);
    blinkState = 1;

    //turn off
  } else if (((currentTime - blinkTimer) > blinkInterval) && (blinkState == 1)) {
    digitalWrite(LED_GREEN, LOW);
    digitalWrite(LED_RED, LOW);
    blinkState = 0;

    //decrement number of blinks
  } else if ((currentTime - blinkTimer) > blinkInterval * 2) {
    blinksRemaining--;
    blinkTimer = millis();
  }
}
