/**
 * @file spu001.c
 * @author Mathieu Le Goc (mathieu.legoc@femtosense.ai)
 * @brief Library providing basic functionalities to implement SPU applications
 * @version 1.1.2
 * @date 2024-03-18
 *
 * @copyright Copyright (c) 2024
 *
 */
#include "spu001.h"
#include "Arduino.h"
#include "evk2_spi.h"
#include "spu001_constants.h"

static const SpuConfiguration kSpuDefaultConfig = {.clock_type = kSpuOsc,
                                                   .interrupt_type = kSpuInterruptPin,
                                                   .crystal_frequency_hz = 32768,
                                                   .core_clock_frequency_mhz = 100,
                                                   .spi_clock_frequency_mhz = 10};
static SpuConfiguration spu_config;
static SpuVersion spu_hw_version = kSpuVersionNone;

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Hardware wrapping functions                                               //
//                                                                           //
// Replace the current HAL calls to match the current SPU host               //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

/**
 * @brief          Local wrapper to read the SPU interrupt GPIO
 *
 * @return         None
 */
static bool SpuInterruptPinRead(void) {
  return digitalRead(SPU_INT_OR_REFCLK) > 0;
}

/**
 * @brief          Local wrapper for microsecond block delay function
 *
 * @param          delay_us Time to wait in microseconds
 * @return         None
 */
static void SpuDelayUs(uint32_t delay_us) {
  delayMicroseconds(delay_us);
}

/**
 * @brief          Local wrapper to enable (drive high) the SPU reset GPIO
 *
 * @return         None
 */
static void SpuResetEnable(void) {
  digitalWrite(SPU_RESET_PIN, HIGH);
}

/**
 * @brief          Local wrapper to disable (drive low) the SPU reset GPIO
 *
 * @return         None
 */
static void SpuResetDisable(void) {
  digitalWrite(SPU_RESET_PIN, LOW);
}

/**
 * @brief          Local wrapper to enable (drive low) the SPU's SPI CS GPIO
 *
 * @return         None
 */
static void SpuSpiEnable(void) {
  digitalWrite(SS_SPU, LOW);
}

/**
 * @brief          Local wrapper to disable (drive high) the SPU's SPI CS GPIO
 *
 * @return         None
 */
static void SpuSpiDisable(void) {
  digitalWrite(SS_SPU, HIGH);
}

/**
 * @brief         Local wrapper for SPI write operation
 *
 * @param         tx_data Pointer to the data to be written
 * @param         tx_length The number of bytes to be written
 * @return        @ref SpuResult
 */
static SpuResult SpuSpiWrite(uint8_t *tx_data, uint16_t tx_length) {
  Evk2SpiWrite(spu_config.spi_clock_frequency_mhz, tx_data, tx_length);
  //always return success for now
  return kSpuSuccess;
}

/*
 *                NOTE ON SPU_READ DEPENDENCE ON THE HOST HAL
 *
 *                If the host SPI HAL supports full-duplex transactions, make sure
 *                SPU_SPI_FULL_DUPLEX is set to 1 in spu001.h. This will result in rx_buffer
 *                containing all the bytes received during the transaction:
 *                ++-------++-----------+-----------++
 *                || bytes || tx_length | rx_length ||
 *                ++-------++-----------+-----------++
 *                ||  mosi || tx_buffer |     0     ||
 *                ++-------++-----------+-----------++
 *                ||  miso ||       rx_buffer       ||
 *                ++-------++-----------------------++
 *
 *                If the host SPI HAL supports only half-duplex transaction, make sure
 *                SPU_SPI_FULL_DUPLEX is set to 0 in spu001.h. This will result in rx_buffer
 *                containing only the data received after tx_buffer:
 *                ++-------++-----------+-----------++
 *                || bytes || tx_length | rx_length ||
 *                ++-------++-----------+-----------++
 *                ||  mosi || tx_buffer |     0     ||
 *                ++-------++-----------+-----------++
 *                ||  miso ||     X     | rx_buffer ||
 *                ++-------++-----------+-----------++
 *
 *                This ensures that the functions implementing SpuSpiRead know how to handle the
 *                data stored in data_buffer.
 */

/**
 * @brief         Local wrapper for SPI read operation
 *
 * @param         tx_buffer Pointer to the buffer containing the read commands
 * @param         tx_length Length of the commands in bytes
 * @param         rx_buffer Pointer to the buffer to store the data read
 * @param         rx_length Length of data to read
 * @return        @ref SpuResult
 */
static SpuResult SpuSpiRead(uint8_t *tx_buffer, uint16_t tx_length, uint8_t *rx_buffer,
                            uint16_t rx_length) {
  Evk2SpiRead(spu_config.spi_clock_frequency_mhz,tx_buffer, tx_length, rx_buffer, rx_length);
  //always return success for now
  return kSpuSuccess;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

/**
 * @brief         Resets the SPU
 * @param         None
 * @return        None
 */
SpuResult SpuReset(void) {
    SpuResetEnable();
    SpuDelayUs(1000);
    SpuResetDisable();
    return kSpuSuccess;
}

/**
 * @brief         Reads the data in registers at the address provided
 *
 * @param         reg_address The base address in memory to read
 * @param         data The pointer to the buffer where the data will be stored
 * @param         length The number of registers to read
 * @return        @ref SpuResult
 */
static SpuResult SpuReadRegisters(SpuAddress reg_address, SpuConfigData *data, uint16_t length) {
    SpuResult result = kSpuSuccess;
    SpuConfigRxMessage rx_message = {0};
    SpuHeader config_header = {.command = kSpiIoReg << 1 | kSpuSpiRead,
                               .address_start = reg_address,
                               .address_end = reg_address + length * sizeof(SpuConfigData)};

    if (data == NULL || length == 0) {
        return kSpuBadParameters;
    }

    SpuSpiEnable();
#if SPU_SPI_FULL_DUPLEX == 1
    result = SpuSpiRead((uint8_t *)&config_header, sizeof(SpuHeader) + kSpuPaddingRxLen,
                        rx_message.data, length * sizeof(SpuConfigData));
#else
    result = SpuSpiRead((uint8_t *)&config_header, sizeof(SpuHeader) + kSpuPaddingRxLen,
                        (uint8_t *)rx_message.message.payload, length * sizeof(SpuConfigData));
#endif
    SpuSpiDisable();

    if (result == kSpuSuccess) {
        memcpy(data, rx_message.message.payload, length * sizeof(SpuConfigData));
        return kSpuSuccess;
    } else {
        return kSpuError;
    }
}

/**
 * @brief         Reads a value in the register at the address provided
 *
 * @param         reg_address The address in memory to read
 * @param         reg_value The pointer to the variable where the data will be stored
 * @return        @ref SpuResult
 */
static SpuResult SpuReadRegister(SpuAddress reg_address, SpuConfigData *reg_value) {
    if (reg_value == NULL) {
        return kSpuBadParameters;
    }
    return SpuReadRegisters(reg_address, reg_value, 1);
}

/**
 * @brief         Writes the data in registers at the address provided
 *
 * @param         reg_address The base address in memory to write
 * @param         data The pointer to the buffer of data to write
 * @param         length The number of registers to write
 * @return        @ref SpuResult
 */
static SpuResult SpuWriteRegisters(SpuAddress reg_address, SpuConfigData *data, uint16_t length) {
    SpuResult result = kSpuSuccess;

    if (data == NULL || length == 0) {
        return kSpuBadParameters;
    }

    SpuConfigTxMessage config_buffer = {
        .message.header.command = kSpiIoReg << 1 | kSpuSpiWrite,
        .message.header.address_start = reg_address,
        .message.header.address_end = reg_address + length * sizeof(SpuConfigData)};

    memcpy(config_buffer.message.payload, data, length * sizeof(SpuConfigData));

    SpuSpiEnable();
    result = SpuSpiWrite(config_buffer.data,
                         sizeof(SpuHeader) + length * sizeof(SpuConfigData) + kSpuPaddingTxLen);
    SpuSpiDisable();

    return result;
}

/**
 * @brief         Writes a value in the register at the address provided
 *
 * @param         reg_address The address in memory to write
 * @param         reg_value The value to write
 * @return        @ref SpuResult
 */
static SpuResult SpuWriteRegister(SpuAddress reg_address, SpuConfigData reg_value) {
    return SpuWriteRegisters(reg_address, &reg_value, 1);
}

/**
 * @brief         Initiate the communication with SPU and checks the values
 * 				  of registers against defaults to confirm the connection
 *
 * @param         None
 * @return        @ref SpuResult
 */
static SpuResult SpuCheckIntegrity(void) {
    SpuConfigData buffer[7];

    if (SpuReadRegisters(kSpuAddrSpiConf, buffer, 7) != kSpuSuccess) {
        return kSpuError;
    }

    if (memcmp(buffer, defaults_reg_tc2, sizeof(buffer)) == 0) {
        spu_hw_version = kSpuVersionTc2;
        return kSpuSuccess;
    } else if (memcmp(buffer, defaults_reg_mp, sizeof(buffer)) == 0) {
        spu_hw_version = kSpuVersionMp;
        return kSpuSuccess;
    } else {
        return kSpuNotDetected;
    }
}

/**
 * @brief         Waits for the PLL to lock
 *
 * @param         timeout_ms Maximum time to wait for the PLL to lock (in ms)
 * @return        @ref SpuResult
 */
static SpuResult SpuWaitPllLock(uint32_t timeout_ms) {
    uint32_t pll_timeout_ms = 0, buffer = 0;

    if (timeout_ms <= 0) {
        return kSpuBadParameters;
    }

    // Poll the register until the bit is set indicating the PLL is locked
    do {
        SpuDelayUs(2000);
        if (SpuReadRegister(kSpuAddrPllLock, &buffer) != kSpuSuccess) {
            return kSpuError;
        }
        pll_timeout_ms += 2;
    } while (!(buffer == 0x01) && pll_timeout_ms < timeout_ms);

    return pll_timeout_ms < timeout_ms ? kSpuSuccess : kSpuPllConfigError;
}

/**
 * @brief Enables or disables the interrupt pin (SPU001MP only)
 *
 * @return SpuResult
 */
static SpuResult SpuEnableInterrupPin(bool enable) {
    SpuConfigData osc_config = 0;
    if (spu_hw_version != kSpuVersionMp) {
        return kSpuSuccess;
    }
    if (SpuReadRegister(kSpuAddrIoConf, &osc_config) != kSpuSuccess) {
        return kSpuConfigError;
    }
    if (enable) {
        osc_config |= kSpuRegIoIntEnable;
    } else {
        osc_config &= kSpuRegIoIntDisable;
    }
    if (SpuWriteRegister(kSpuAddrIoConf, osc_config) != kSpuSuccess) {
        return kSpuConfigError;
    }
    return kSpuSuccess;
}

/**
 * @brief Enables the SPU oscillator
 *
 * @param clk_source Type of clock to use
 * @return SpuResult
 */
static SpuResult SpuEnableOscillator(SpuClockType clk_source) {
    SpuConfigData osc_config = 0;
    osc_config = (clk_source == kSpuOsc) ? kSpuRegIoConfOsc : kSpuRegIoConfExtClk;

    if (SpuWriteRegister(kSpuAddrIoConf, osc_config) != kSpuSuccess) {
        return kSpuConfigError;
    }

    if (SpuEnableInterrupPin(spu_config.interrupt_type == kSpuInterruptPin) != kSpuSuccess) {
        return kSpuConfigError;
    }

    return kSpuSuccess;
}

/**
 * @brief Disables the SPU's oscillator
 *
 * @return SpuResult
 */
static SpuResult SpuDisableOscillator(void) {
    return SpuWriteRegister(kSpuAddrIoConf, kSpuRegIoConfNoOsc);
}

/**
 * @brief         Configure SPU to use the desired clock
 *
 * @param         core_frequency_mhz  Clock frequency to run the SPU in MHz
 * @param         clk_source Clock source to use
 * @param         clk_frequency_hz Frequency of the clock source in Hz
 * @return        @ref SpuResult
 */
static SpuResult SpuConfigureClock(uint32_t core_frequency_mhz, SpuClockType clk_source,
                                   uint32_t clk_frequency_hz) {
    SpuClockRegisters clock_reg = {
        .registers.clkr = 0,
        .registers.clkod = 0,
        .registers.clkf = (core_frequency_mhz * 1000000 / clk_frequency_hz - 1),
        .registers.bwadj = ((core_frequency_mhz * 1000000) / (clk_frequency_hz * 2) - 1)};

    // Hardware bug in SPU001 TC2 requires these to be shifted
    if (spu_hw_version == kSpuVersionTc2) {
        clock_reg.registers.clkf <<= 1;
        clock_reg.registers.bwadj <<= 1;
    }
    //  Write the PLL configuration registers
    if (SpuWriteRegisters(kSpuAddrPllConf1, clock_reg.data,
                          sizeof(SpuClockRegisters) / sizeof(SpuConfigData)) != kSpuSuccess) {
        return kSpuConfigError;
    }

    // For SPU001 MP, turn on the Cores Reset signal during configuration
    if (spu_hw_version == kSpuVersionMp) {
        if (SpuWriteRegister(kSpuAddrSpiConf, kSpuRegSpiConfCoresReset |
                                                  kSpuRegSpiConfCore0Disable |
                                                  kSpuRegSpiConfCore1Disable) != kSpuSuccess) {
            return kSpuConfigError;
        }
    }
    if (SpuEnablePll(clk_source) != kSpuSuccess) {
        return kSpuConfigError;
    }
    // End of configuration sequence for SPU001 TC2, return at this point
    if (spu_hw_version == kSpuVersionTc2) {
        return kSpuSuccess;
    }
    // Configuration continues for SPU001 MP
    // Enable both cores and disable the cores reset
    if (SpuWriteRegister(kSpuAddrSpiConf, 0) != kSpuSuccess) {
        return kSpuConfigError;
    }
    return kSpuSuccess;
}

/**
 * @brief         Flashes SPU with the identity model (acts as a loopback and returns exactly what
 *                came in) and checks with several batches of 64 samples that the communication
 *                runs as expected
 *
 * @param         None
 * @return        @ref SpuResult
 */
static SpuResult SpuTestAudioStream(void) {
    SpuData test_tx[16] = {0}, test_rx[16] = {0};
    uint8_t batch_id = 1;

    // flash the SPU with the identity model
    if (spu_hw_version == kSpuVersionTc2) {
        for (uint32_t i = 0; i < sizeof(kIdentityAddressesTc2) / sizeof(SpuConfigData); i++) {
            if (SpuWriteModelValue(kIdentityAddressesTc2[i], kIdentityDataTc2[i]) != kSpuSuccess) {
                return kSpuConfigError;
            }
        }
    } else if (spu_hw_version == kSpuVersionMp) {
        for (uint32_t i = 0; i < sizeof(kIdentityAddressesMp) / sizeof(SpuConfigData); i++) {
            if (SpuWriteModelValue(kIdentityAddressesMp[i], kIdentityDataMp[i]) != kSpuSuccess) {
                return kSpuConfigError;
            }
        }
    }
    SpuDelayUs(5000);
    while (batch_id <= 2) {
        // generate arbitrary data
        for (uint8_t i = 0; i < 16; i++) {
            test_tx[i] = batch_id << 8 | i;
        }
        // send it to SPU
        if (SpuWriteData(test_tx, 16, 1, 1) != kSpuSuccess) {
            return kSpuError;
        }
        SpuDelayUs(5000);
        // check that SPU is done processing
        if (SpuPollInterrupt() != kSpuSuccess) {
            return kSpuNotReady;
        }
        if (SpuReadData(test_rx, 16) != kSpuSuccess) {
            return kSpuError;
        }
        // check that the values read match the values sent
        if (memcmp(test_rx, test_tx, sizeof(test_rx)) != 0) {
            return kSpuTestFailed;
        }
        batch_id++;
    }
    return kSpuSuccess;
}

/**
 * @brief Sets the SPU clock output divider to reduce core_frequency_mhz set in
 *        SpuConfigureClock following this equation:
 *        effective_core_frequency = core_frequency_mhz / divider
 *        effective_core_frequency should be > 2 * SPI_CLK_FREQUENCY to maintain SPI communication
 *        capabilities.
 *
 * @param divider Value of the output divider (1 or even value between 2 and 16)
 *                divider
 * @return SpuResult
 */
// TODO: make this function static
SpuResult SpuClockSetOutputDivider(uint8_t divider) {
    // Make sure divider is an even number by clearing the LSB
    SpuConfigData clock_od = divider > 1 ? (divider & (UINT32_MAX - 1)) : 1;
    // clock_od should be [0,1,3,5,7,9,11,13,15]
    clock_od = MIN(clock_od, (uint32_t)16) - 1;
    if (spu_hw_version == kSpuVersionTc2) {
        clock_od <<= 1;
    }
    if (SpuWriteRegister(kSpuAddrPllConf2, clock_od) != kSpuSuccess) {
        return kSpuConfigError;
    }
    return kSpuSuccess;
}

/**
 * @brief         Control the SPU cores
 *
 * @param         core_0 Enable core #0
 * @param         core_1 Enable core #1
 * @return        @ref SpuResult
 */
SpuResult SpuEnableCores(bool core_0, bool core_1) {
    SpuConfigData cores_config =
        (core_0 ? 0 : kSpuRegSpiConfCore0Disable) | (core_1 ? 0 : kSpuRegSpiConfCore1Disable);

    if (SpuWriteRegister(kSpuAddrSpiConf, cores_config) != kSpuSuccess) {
        return kSpuConfigError;
    }
    return kSpuSuccess;
}

/**
 * @brief Enables the SPU PLL
 *
 * @return SpuResult
 */
SpuResult SpuEnablePll(SpuClockType clk_source) {
    if (SpuEnableOscillator(clk_source) != kSpuSuccess) {
        return kSpuConfigError;
    }
    // Reset PLL to update, turn it off
    if (SpuWriteRegister(kSpuAddrPllConf0, 1) != kSpuSuccess) {
        return kSpuConfigError;
    }
    // and back on
    if (SpuWriteRegister(kSpuAddrPllConf0, 0) != kSpuSuccess) {
        return kSpuConfigError;
    }
    // Wait for the PLL to lock
    if (SpuWaitPllLock(1000) != kSpuSuccess) {
        return kSpuTimeout;
    }
      // End of configuration sequence for SPU001 TC2, return at this point
    if (spu_hw_version == kSpuVersionTc2) {
      return kSpuSuccess;
    }

    // Configuration continues for SPU001 MP
    // Ungate the PLL output
    if (SpuWriteRegister(kSpuAddrPllConf0, kSpuRegPll0ConfClockEnable) != kSpuSuccess) {
      return kSpuConfigError;
    }
  
    return kSpuSuccess;
}

/**
 * @brief Disables the SPU PLL
 *
 * @return SpuResult
 */
SpuResult SpuDisablePll(void) {
    if (SpuDisableOscillator() != kSpuSuccess) {
        return kSpuConfigError;
    }
    if (SpuWriteRegister(kSpuAddrPllConf0, 0x20) != kSpuSuccess) {
        return kSpuConfigError;
    }

    return kSpuSuccess;
}

/**
 * @brief         Write a model value in SPU's memory
 *
 * @param         address The base address in memory to write
 * @param         data The model value to write
 * @return        @ref SpuResult
 */
SpuResult SpuWriteModelValue(SpuAddress address, SpuModelData data) {
    SpuResult result = kSpuSuccess;

    SpuConfigTxMessage config_buffer = {
        .message.header.command = kSpiApb << 1 | kSpuSpiWrite,
        .message.header.address_start = address,
        .message.header.address_end = address + sizeof(SpuModelData)};

    memcpy(config_buffer.message.payload, &data, sizeof(SpuModelData));

    SpuSpiEnable();
    result = SpuSpiWrite(config_buffer.data,
                         sizeof(SpuHeader) + sizeof(SpuModelData) + kSpuPaddingTxLen);
    SpuSpiDisable();

    return result;
}

/**
 * @brief         Write a batch of model values in SPU's memory
 *
 * @param         address The base address in memory to write
 * @param         data The pointer to the model buffer to write
 * @param         length The pointer to the model buffer to write
 * @return        @ref SpuResult
 */
SpuResult SpuWriteModelBatch(SpuAddress address, SpuModelData *data, uint16_t length) {
    SpuResult result = kSpuSuccess;
    SpuConfigTxMessage config_buffer = {
        .message.header.command = kSpiApb << 1 | kSpuSpiWrite,
        .message.header.address_start = address,
        .message.header.address_end = address + length * sizeof(SpuModelData)};

    if (length == 0 || data == NULL) {
        return kSpuBadParameters;
    }
    memcpy(config_buffer.message.payload, data, length * sizeof(SpuModelData));

    SpuSpiEnable();
    result = SpuSpiWrite(config_buffer.data,
                         sizeof(SpuHeader) + length * sizeof(SpuModelData) + kSpuPaddingTxLen);
    SpuSpiDisable();

    return result;
}

/**
 * @brief         Polls the SPU interrupt register
 *
 * @param         None
 * @return        @ref SpuResult
 */
static SpuResult SpuPollInterruptRegister(void) {
    SpuConfigData interrupt_reg = 0;
    if (SpuReadRegister(kSpuAddrInterrupt, &interrupt_reg) != kSpuSuccess) {
        return kSpuError;
    }
    return interrupt_reg > 0 ? kSpuSuccess : kSpuNotReady;
}

/**
 * @brief         Polls the SPU interrupt pin while clocking the SPI with one transaction
 *                (necessary as SPU INT is clock synchronous)
 *
 * @param         None
 * @return        @ref SpuResult
 */
static SpuResult SpuPollInterruptPin(void) {
    uint8_t dummy_bytes[4] = {0};

    if (SpuInterruptPinRead() == 1) {
        return kSpuSuccess;
    } else {
        // SPU001 TC2's interrupt is synchronous and requires clock edges on SPI_CLK to update
        if (spu_hw_version == kSpuVersionTc2) {
            SpuSpiWrite(dummy_bytes, sizeof(dummy_bytes));
            return SpuInterruptPinRead() == 1 ? kSpuSuccess : kSpuNotReady;
        }
        return kSpuNotReady;
    }
}

/**
 * @brief         Polls the SPU interrupt following the interrupt type selected in the
 * configuration
 *
 * @return        @ref SpuResult
 */
SpuResult SpuPollInterrupt(void) {
    return (spu_config.interrupt_type == kSpuInterruptPin) ? SpuPollInterruptPin()
                                                           : SpuPollInterruptRegister();
}

/**
 * @brief         Writes data to be processed by the SPU. Every transaction must have at least one first and last message. These may be the same message.
 *
 * @param         length The number of samples to write
 * @param         data The pointer to the buffer of audio data to write
 * @param         first_message A header is included in the first message
 * @param         last_message Processing begins after the last message is sent
 * @return        @ref SpuResult
 */
SpuResult SpuWriteData(SpuData *data, uint16_t length, bool first_message, bool last_message) {
    SpuResult result = kSpuSuccess;
    
    if (data == NULL || length == 0) {
        return kSpuBadParameters;
    }
    
    //The first message includes a header
    if (first_message) {
      SpuDataTxMessage data_buffer = {
          .message.header.command = (last_message ? kSpiAxisLast : kSpiAxis) << 1 | kSpuSpiWrite,
          .message.header.address_start = kSpuAddrAudio,
          .message.header.address_end =
              kSpuAddrAudio + sizeof(SpuInputParams) + length * sizeof(SpuData),
          .message.params.core_id = 0,
          .message.params.program_counter = 0};

      memcpy(data_buffer.message.payload, data, length * sizeof(SpuData));

      SpuSpiEnable();
      result = SpuSpiWrite(data_buffer.data, sizeof(SpuHeader) + sizeof(SpuInputParams) +
                                                length * sizeof(SpuData) + kSpuPaddingTxLen);
      SpuSpiDisable();
    } else {
      SpuDataTxMessageContinued data_buffer = {
          .message.header.command = (last_message ? kSpiAxisLast : kSpiAxis) << 1 | kSpuSpiWrite,
          .message.header.address_start = kSpuAddrAudio,
          .message.header.address_end =
              kSpuAddrAudio + length * sizeof(SpuData)};

      memcpy(data_buffer.message.payload, data, length * sizeof(SpuData));

      SpuSpiEnable();
      result = SpuSpiWrite(data_buffer.data, sizeof(SpuHeader) + length * sizeof(SpuData) +
                                                 kSpuPaddingTxLen);
      SpuSpiDisable();
    }

    return result;
}

/**
 * @brief         Reads data processed by the SPU
 *
 * @param         data The pointer to the buffer where the audio data to be stored
 * @param         length The number of samples to read
 * @return        @ref SpuResult
 */
SpuResult SpuReadData(SpuData *data, uint16_t length) {
    SpuResult result = kSpuSuccess;
    SpuDataRxMessage rx_message = {0};
    SpuHeader data_header = {
        .command = kSpiAxisLast << 1 | kSpuSpiRead,
        .address_start = kSpuAddrAudio,
        .address_end = kSpuAddrAudio + length * sizeof(SpuData) + sizeof(SpuOutputParams)};

    if (data == NULL || length == 0) {
        return kSpuBadParameters;
    }
    SpuSpiEnable();
#if SPU_SPI_FULL_DUPLEX == 1
    result = SpuSpiRead((uint8_t *)&data_header, sizeof(SpuHeader) + kSpuPaddingRxLen,
                        rx_message.data, sizeof(SpuOutputParams) + length * sizeof(SpuData));
#else
    result = SpuSpiRead((uint8_t *)&data_header, sizeof(SpuHeader) + kSpuPaddingRxLen,
                        (uint8_t *)&rx_message.message.params,
                        sizeof(SpuOutputParams) + length * sizeof(SpuData));
#endif
    SpuSpiDisable();

    if (result == kSpuSuccess && rx_message.message.params.mailbox_id == kSpuMailboxIdValid &&
        rx_message.message.params.core_id == kSpuCoreIdValid) {
        memcpy(data, rx_message.message.payload, length * sizeof(SpuData));
        return kSpuSuccess;
    } else {
        return kSpuFrameError;
    }
}

/**
 * @brief Turns on the SPU sleep mode
 *
 * @return SpuResult
 */
SpuResult SpuEnterSleepMode(void) {
    return SpuClockSetOutputDivider(spu_config.core_clock_frequency_mhz /
                                    (2 * spu_config.spi_clock_frequency_mhz));
}

/**
 * @brief Turns on the SPU sleep mode
 *
 * @return SpuResult
 */
SpuResult SpuExitSleepMode(void) { return SpuClockSetOutputDivider(1); }

/**
 * @brief         Initialize the SPU and its dependencies (SPI and GPIOs)
 *
 * @param         config Pointer to the structure containing the SPU configuration parameters
 * @return        @ref SpuResult
 */
SpuResult SpuInit(SpuConfiguration *config) {
    SpuResult result;
    SpuReset();
    if ((result = SpuCheckIntegrity()) == kSpuNotDetected) {
        return result;
    }
    // Store the configuration
    if (config == NULL) {
        memcpy(&spu_config, &kSpuDefaultConfig, sizeof(SpuConfiguration));
    } else {
        memcpy(&spu_config, config, sizeof(SpuConfiguration));
        // SPU001 TC2 supports only pin-based interrupts
        if (spu_hw_version == kSpuVersionTc2) {
            spu_config.interrupt_type = kSpuInterruptPin;
        }
    }

    if ((result = SpuConfigureClock(spu_config.core_clock_frequency_mhz, spu_config.clock_type,
                                    spu_config.crystal_frequency_hz)) != kSpuSuccess) {
        return result;
    }
    if (SpuEnableInterrupPin(spu_config.interrupt_type == kSpuInterruptPin) != kSpuSuccess) {
        return kSpuConfigError;
    }
    if ((result = SpuTestAudioStream()) != kSpuSuccess) {
        // Repeat the reset sequence in case SPU is configured but still not running normally
        // (can happen after a crash. In such case SPU must be configured and the PLL locked,
        // SPU then undergoes a hard reset and a complete configuration
        SpuReset();

        if ((result = SpuCheckIntegrity()) == kSpuNotDetected) {
            return result;
        }
        if ((result = SpuConfigureClock(spu_config.core_clock_frequency_mhz, spu_config.clock_type,
                                        spu_config.crystal_frequency_hz)) != kSpuSuccess) {
            return result;
        }
        if (SpuEnableInterrupPin(spu_config.interrupt_type == kSpuInterruptPin) != kSpuSuccess) {
            return kSpuConfigError;
        }
        return SpuTestAudioStream();
    }
    return kSpuSuccess;
}

/**
 * @brief Provides a description for the error token provided
 *
 * @param error
 * @return const char*
 */
const char *SpuResultToString(SpuResult error) {
    const char *descriptions[] = {"Success",      "Error",
                                  "Not detected", "Configuration error",
                                  "Frame error",  "Bad parameters",
                                  "Not ready",    "Test failed",
                                  "Timeout",      "PLL configuration failed"};
    return descriptions[error];
}

/**
 * @brief Returns the hardware version of SPU001 once detected and initialized
 *
 * @return SpuVersion
 */
SpuVersion SpuGetHardwareVersion(void) { return spu_hw_version; }
